
#pragma once

#include <vector>
#include <algorithm>

// for debug
#ifdef _DEBUG
	const bool _Mtl_ExplorerMenu_traceOn = false;
	#define epmTRACE if (_Mtl_ExplorerMenu_traceOn) ATLTRACE
#else
	#define epmTRACE
#endif

// be care, if no element, &array[0] is invalid.
template <class _Key, class _Val>
class CCrasyMap : public CSimpleArray< std::pair<_Key, _Val> >
{
public:
	bool FindKey(const _Key& __key)
	{
		if (GetSize() == 0)
			return false;

		return end() != std::find_if(begin(), end(), _Finder(__key));
	}

	void Add(const _Key& __key, const _Val& __val)
	{
		if (GetSize() == 0) {
			CSimpleArray< std::pair<_Key, _Val> >::Add(std::make_pair(__key, __val));
			return;
		}

		std::pair<_Key, _Val>* p = std::find_if(begin(), end(), _Finder(__key));
		if (p == end())
			CSimpleArray< std::pair<_Key, _Val> >::Add(std::make_pair(__key, __val));
		else
			(*p).second = __val;
	}

	_Val Lookup(const _Key& __key)
	{
		if (GetSize() == 0)
			return _Val();

		std::pair<_Key, _Val>* p = std::find_if(begin(), end(), _Finder(__key));
		if (p != end())
			return p->second;
		else
			return _Val();
	}

	void Sort()
	{
		if (GetSize() == 0)
			return;

		std::sort(begin(), end(), _Compare());
	}

	std::pair<_Key, _Val>* begin()
	{
		return &(*this)[0];
	}

	std::pair<_Key, _Val>* end()
	{
		return &(*this)[0] + GetSize();
	}

//	template <class _Key, class _Val>
	struct _Finder
	{
		_Key __aKey;
		_Finder(_Key __key) : __aKey(__key) { }
		bool operator()(const std::pair<_Key, _Val>& src)
		{
			return (__aKey == src.first);
		}
	};

//	template <class _Key, class _Val>
	struct _Compare
	{
		bool operator()(const std::pair<_Key, _Val>& x, const std::pair<_Key, _Val>& y)
		{
			return x.second < y.second;
		}
	};
};

// Extended styles
#define EMS_ADDITIONALMENUITEM		0x00000001L
#define EMS_ADDITIONALMENUITEMNOSEP 0x00000002L
#define EMS_USER					0x00010000L

#define PASS_MSG_MAP_MENUOWNER_TO_EXPMENU(x) CHAIN_MSG_MAP_MEMBER(x)
/*
class ATL_NO_VTABLE CExplorerMenuImplBase
{
public:
	static int s_nMaxMenuItemTextLength;
	static int s_nMaxMenuBreakCount;

	static void SetMaxMenuItemTextLength(int n)
	{
		s_nMaxMenuItemTextLength = n;
	}

	static int GetMaxMenuItemTextLength()
	{
		return s_nMaxMenuItemTextLength;
	}

	static void SetMaxMenuBreakCount(int n)
	{
		s_nMaxMenuBreakCount = n;
	}

	static int GetMaxMenuBreakCount()
	{
		return s_nMaxMenuBreakCount;
	}
};

__declspec(selectany) int CExplorerMenuImplBase::s_nMaxMenuItemTextLength = 50;
__declspec(selectany) int CExplorerMenuImplBase::s_nMaxMenuBreakCount = 5000;
*/

template <class T>
class ATL_NO_VTABLE CExplorerMenuImpl// : public CExplorerMenuImplBase
{
public:
// Declarations

// Constants

// Data members
	DWORD m_dwStyle;

	CMenuHandle m_menu;
//	CMenu m_menuOriginal;
	int m_nOriginalMenuItemCountExceptInsertPointMenuItem;
	int m_nOriginalInsertPoint;
	int m_nMaxMenuBreakCount;
	int m_nMaxMenuItemTextLength;

	CString m_strRootDirectoryPath;
	CString m_strNoEntries;
	CString m_strAdditional;

	CSimpleArray<HMENU> m_arrMenuHandle;
	int m_nCurrentMenuID;
	CCrasyMap<int, CString> m_mapID;			// id and file path
	CCrasyMap<HMENU, CString> m_mapSubMenu; 	// menu handle and file path

	int m_nInsertPointMenuItemID;
	int m_nMinID, m_nMaxID;

	CSimpleArray<CString> m_arrIgnoredPath;

// Constructor
	CExplorerMenuImpl(int nInsertPointMenuItemID, const CString& strNoEntries = _T("(empty)"),
		int nMinID = 0x7000, int nMaxID = 0x7FFF,
		const CString& strAdditional = _T("Open This Directory"), int nMaxMenuItemTextLength = 55)
		: m_strNoEntries(strNoEntries), m_strAdditional(strAdditional),
		  m_nInsertPointMenuItemID(nInsertPointMenuItemID),
		  m_nMinID(nMinID), m_nMaxID(nMaxID),
		  m_dwStyle(0), m_nMaxMenuBreakCount(5000), m_nMaxMenuItemTextLength(55)
	{
		m_nCurrentMenuID = m_nMinID;
	}

	BOOL InstallExplorerMenu(HMENU hMenu)
	{
		SetMenuHandle(hMenu);
		return TRUE;
	}

	~CExplorerMenuImpl()
	{
	}

// Overridables
	void OnMenuItemInitialUpdate(const CString& strDirPath, CSimpleArray<CMenuItemInfo>& infos)
	{
		std::sort(_begin(infos), _end(infos), _DefaultMenuItemCompare(*this));
	}

	void OnExecute(const CString& strFilePath)
	{
	}

	CString OnGetMenuItemText(const CString& strPath)
	{
		CString str = MtlGetDisplayTextFromPath(strPath);
		str = MtlCompactString(str, m_nMaxMenuItemTextLength);
		str.Replace(_T("&"), _T("&&"));

		return str;
	}

// Attributes
	void SetMenuHandle(HMENU hMenu)
	{
		ATLASSERT(::IsMenu(hMenu));

		// save original menu infos
		ATLASSERT(_FindInsertPoint(hMenu) != -1);// need InsertPointMenuItem
		m_nOriginalMenuItemCountExceptInsertPointMenuItem = ::GetMenuItemCount(hMenu) - 1;
		m_nOriginalInsertPoint = _FindInsertPoint(hMenu);

		m_menu = hMenu;
	}

	CString GetRootDirectoryPath() const
	{
		return m_strRootDirectoryPath;
	}

	void SetRootDirectoryPath(const CString& strDirectoryPath)
	{
//		ATLASSERT(strDirectoryPath.IsEmpty() == FALSE);
		m_strRootDirectoryPath = strDirectoryPath;
	}

	BOOL LoadAdditionalMenu(_U_STRINGorID menu)
	{
		return m_menuAdditional.LoadMenu(menu);
	}

	void SetStyle(DWORD dwStyle)
	{
		m_dwStyle = dwStyle;
	}

	DWORD GetStyle() const
	{
		return m_dwStyle;
	}

	void AddIgnoredPath(CString& strPath)
	{
		m_arrIgnoredPath.Add(strPath);
	}

	void ResetIgnoredPath()
	{
		m_arrIgnoredPath.RemoveAll();
	}

// Methods
/*
	template <class _Profile>
	void GetProfileExpMenu(_Profile& __profile)
	{
		DWORD dwStyle, dwLength, dwCount;
		LONG lRet = __profile.QueryValue(dwStyle, _T("Style"));
		if (lRet != ERROR_SUCCESS)
			dwStyle = 0;

		lRet = __profile.QueryValue(dwLength,  _T("Max_Text_Length"));
		if (lRet != ERROR_SUCCESS)
			dwLength = 55;

		lRet = __profile.QueryValue(dwCount,  _T("Max_Break_Count"));
		if (lRet != ERROR_SUCCESS)
			dwCount = 5000;

		SetStyle(dwStyle);
		SetMaxMenuItemTextLength(dwLength);
		SetMaxMenuBreakCount(dwCount);
	}

	template <class _Profile>
	void WriteProfileExpMenu(_Profile& __profile)
	{
		__profile.SetValue(GetStyle(), _T("Style"));
		__profile.SetValue(GetMaxMenuItemTextLength(), _T("Max_Text_Length"));
		__profile.SetValue(GetMaxMenuBreakCount(), _T("Max_Break_Count"));
	}
*/


// Message map and handlers
	BEGIN_MSG_MAP(CExplorerMenu)
		MSG_WM_INITMENUPOPUP(OnInitMenuPopup)
		MESSAGE_HANDLER(WM_MENUSELECT, OnMenuSelect)
		COMMAND_RANGE_HANDLER_EX(m_nMinID, m_nMaxID, OnCommandRange)
	END_MSG_MAP()

	void OnInitMenuPopup(HMENU hmenuPopup, UINT uPos, BOOL fSystemMenu)
	{
		SetMsgHandled(FALSE);
		if (m_strRootDirectoryPath.IsEmpty())
			return;

		epmTRACE(_T("CExplorerMenu::OnInitMenuPopup : %d\n"), m_nCurrentMenuID);

//		Somehow, fSystemMenu is sometimes wrong value, why?
//		if (fSystemMenu) {// window menu, do nothing
//			epmTRACE(_T(" It's system menu, return");
//			return;
//		}

		if (m_menu.m_hMenu == hmenuPopup) {// root menu
			epmTRACE(_T(" It's root menu\n"));
			// delayed clean up if menu not clicked
			m_mapID.RemoveAll();
			// at first, add it to the list
			m_arrMenuHandle.Add(hmenuPopup);

			// restore original menu (it's need cuz WM_MENUSELECT is not sent if not toplevel menu. for example. chevron menu)
			_RestoreOriginalMenu();

			// delete insert point menu item
			_DeleteInsertPointMenuItem();

			int nInsertPoint = m_nOriginalInsertPoint;
			_Explore(m_menu, nInsertPoint, m_strRootDirectoryPath);
		}
		else if (m_arrMenuHandle.Find(hmenuPopup) != -1){// it's my menu
//			epmTRACE(_T(" It's %s menu\n"), m_strTestText);
			// first, clean up all the menu items
			_DeleteAllMenuItems(hmenuPopup);
			_Explore(hmenuPopup, 0, _LoadDirectoryPath(hmenuPopup));
			SetMsgHandled(TRUE);// eat it!!
		}
	}

	void OnCommandRange(UINT, int nID, HWND hWndCtrl)
	{
		if (hWndCtrl) {
			SetMsgHandled(FALSE);
			return;
		}

		epmTRACE(_T("CExplorerMenu::OnCommandRange : %d\n"), nID);
		T* pT = static_cast<T*>(this);
		if (CString() != m_mapID.Lookup(nID))
			pT->OnExecute(m_mapID.Lookup(nID));
		m_mapID.RemoveAll();// delayed clean up.
	}

	LRESULT OnMenuSelect(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		epmTRACE(_T("CExplorerMenu::OnMenuSelect\n"));
		UINT uItem = LOWORD(wParam);
		UINT fuFlags = HIWORD(wParam);
		HMENU hmenu = (HMENU)lParam;

		SetMsgHandled(FALSE);
		bHandled = FALSE;// must be, oh my god!!!!!!

		// Check if a menu is closing, do a cleanup
		if (fuFlags == 0xFFFF && hmenu == NULL) {// menu closing
			// NOTE: this message sent only if top level menu!!!
			epmTRACE(_T(" Cleanup\n"));
			_CleanUpExceptIDMap();
		}
//		else if (m_arrMenuHandle.Find(hmenu) != -1){// it's my menu
//		}

		return 0;
	}

	void _CleanUpExceptIDMap()
	{
		epmTRACE(_T("CExplorereMenu::_CleanUpExceptIDMap\n"));

		// restore original menu
		_RestoreOriginalMenu();

		m_arrMenuHandle.RemoveAll();
		m_mapSubMenu.RemoveAll();
		m_nCurrentMenuID = m_nMinID;
	}

	void _AddNoneMenuItem(CMenuHandle menu, int& nInsertPoint)
	{
		CMenuItemInfo mii;
		mii.fMask = MIIM_STATE | MIIM_TYPE;
		mii.fType = MFT_STRING;
		mii.fState = MFS_GRAYED;
		mii.dwTypeData = (LPTSTR)(LPCTSTR)m_strNoEntries;
		menu.InsertMenuItem(nInsertPoint++, TRUE, &mii);
	}


	struct _DefaultMenuItemCompare : public std::binary_function<const CMenuItemInfo&, const CMenuItemInfo&, bool>
	{
		CExplorerMenuImpl<T>* _expmenu;

		_DefaultMenuItemCompare(CExplorerMenuImpl<T>& m) : _expmenu(&m)
		{
		}

		bool operator()(const CMenuItemInfo& x, const CMenuItemInfo& y)
		{
			CString strPathA = _expmenu->_GetFilePath(x);
			CString strPathB = _expmenu->_GetFilePath(y);

			bool bDirA = MtlIsDirectoryPath(strPathA);
			bool bDirB = MtlIsDirectoryPath(strPathB);

			if (bDirA == bDirB)
				return strPathA < strPathB;
			else {
				if (bDirA)
					return true;
				else
					return false;
			}
		}
	};

	struct _SetupMenuItemInfoToArray : public std::unary_function<const CString&, void>
	{
		CSimpleArray<CMenuItemInfo>* array;
		CExplorerMenuImpl<T>* expmenu;

		_SetupMenuItemInfoToArray(CSimpleArray<CMenuItemInfo>& a, CExplorerMenuImpl<T>& m)
			: array(&a), expmenu(&m)
		{
		}

		void operator()(const CString& strPath, bool bDir)
		{
			if (bDir){
				CString strDirectoryPath2 = strPath;
				if (expmenu->m_arrIgnoredPath.Find(strDirectoryPath2) != -1)
					return;

				CMenuHandle menuSub;
				menuSub.CreatePopupMenu();

				// add to map and array of menus
				expmenu->_SaveDirectoryPath(menuSub.m_hMenu, strPath);
				expmenu->m_arrMenuHandle.Add(menuSub.m_hMenu);

				LPTSTR lpstrText = new TCHAR[strPath.GetLength() + 1];
				::lstrcpy(lpstrText, strPath);

				CMenuItemInfo mii;
				mii.fMask = MIIM_SUBMENU | MIIM_TYPE;
				mii.fType = MFT_STRING;
				mii.hSubMenu = menuSub.m_hMenu;
				mii.dwTypeData = lpstrText;
				array->Add(mii);
			}
			else {
				CString strFilePath2 = strPath;
				if (expmenu->m_arrIgnoredPath.Find(strFilePath2) != -1)
					return;

				LPTSTR lpstrText = new TCHAR[strPath.GetLength() + 1];
				::lstrcpy(lpstrText, strPath);

				CMenuItemInfo mii;
				mii.fMask = MIIM_ID | MIIM_TYPE;
				mii.fType = MFT_STRING;
				mii.wID = ++expmenu->m_nCurrentMenuID;
				mii.dwTypeData = lpstrText;

				if (expmenu->m_nCurrentMenuID < expmenu->m_nMaxID) {
					// add to command map
					expmenu->m_mapID.Add(expmenu->m_nCurrentMenuID, strPath);
					array->Add(mii);
				}
			}

		}
	};

	void _Explore(CMenuHandle menuPopup, int nInsertPoint, const CString& strDirectoryPath)
	{
		int nOldInsertPoint = nInsertPoint;
		int nBreakCount = nInsertPoint;

		if ((m_dwStyle & EMS_ADDITIONALMENUITEM) && MtlIsFileExist(strDirectoryPath))
			_AddAdditionalMenuItem(menuPopup, nInsertPoint, nBreakCount, strDirectoryPath);

		CSimpleArray<CMenuItemInfo> infos;
		MtlForEachObject_OldShell(strDirectoryPath, _SetupMenuItemInfoToArray(infos, *this));

		// setup menu item info
		T* pT = static_cast<T*>(this);
		pT->OnMenuItemInitialUpdate(strDirectoryPath, infos);

		int i;
		// shorten text size
		for (i = 0; i < infos.GetSize(); ++i) {
			CMenuItemInfo& info = infos[i];
			LPTSTR lpsz = info.dwTypeData;
			CString str = _GetMenuItemText(lpsz);
			delete [] lpsz;
			lpsz = new TCHAR[str.GetLength() + 1];
			::lstrcpy(lpsz, str);
			info.dwTypeData = lpsz;
		}

		// insert menu items
		std::for_each(_begin(infos), _end(infos), _MenuItemInserter(menuPopup, nInsertPoint, nBreakCount, m_nMaxMenuBreakCount));

		// add "none" menu item
		if (infos.GetSize() == 0)
			_AddNoneMenuItem(menuPopup, nInsertPoint);

		// clean up texts
		for (i = 0; i < infos.GetSize(); ++i)
			delete [] infos[i].dwTypeData;
	}

	void _AddAdditionalMenuItem(CMenuHandle menuDest, int& nInsertPoint, int& nBreakCount, const CString& strDirectory)
	{
		CString strDirectoryPath = strDirectory;
		MtlMakeSureTrailingBackSlash(strDirectoryPath);

		CMenuItemInfo mii;
		mii.fMask = MIIM_ID | MIIM_TYPE;
		mii.fType = MFT_STRING;
		mii.wID = ++m_nCurrentMenuID;
		mii.dwTypeData = (LPTSTR)(LPCTSTR)m_strAdditional;
		menuDest.InsertMenuItem(nInsertPoint++, TRUE, &mii);
		++nBreakCount;

		if (m_nCurrentMenuID < m_nMaxID) {
			m_mapID.Add(m_nCurrentMenuID, strDirectoryPath);
		}
		else {
			ATLASSERT(FALSE);
//			::MessageBeep(MB_ICONEXCLAMATION);
		}

		if (!(m_dwStyle & EMS_ADDITIONALMENUITEMNOSEP))
			_InsertSeparator(menuDest, nInsertPoint);
	}

	void _InsertSeparator(CMenuHandle menuDest, int& nInsertPoint)
	{
		CMenuItemInfo mii;
		mii.fMask = MIIM_TYPE;
		mii.fType = MF_SEPARATOR;
		menuDest.InsertMenuItem(nInsertPoint++, TRUE, &mii);
	}

	struct _MenuItemInserter
	{
		CMenuHandle _menu;
		int& _nInsertPoint;
		int& _nBreakCount;
		int _nMaxMenuBreakCount;

		_MenuItemInserter(CMenuHandle menu, int& nInsertPoint, int& nBreakCount, int nMaxMenuBreakCount)
			: _menu(menu), _nInsertPoint(nInsertPoint), _nBreakCount(nBreakCount),
			  _nMaxMenuBreakCount(nMaxMenuBreakCount)
		{
		}

		void operator()(CMenuItemInfo& mii)
		{
			if (_nBreakCount + 1 > _nMaxMenuBreakCount) {// what a difficult algorithm...
				_nBreakCount = 0;// reset
				mii.fType |= MFT_MENUBREAK;
			}

			++_nBreakCount;

			epmTRACE(_T("InsertMenuItem(%s, %d)\n"), mii.dwTypeData, _nInsertPoint);
			_menu.InsertMenuItem(_nInsertPoint++, TRUE, &mii);
		}
	};

	static void _DeleteAllMenuItems(CMenuHandle menu)
	{
		for (int i = 0; i < menu.GetMenuItemCount(); ++i) {
			CMenuHandle menuSub = menu.GetSubMenu(i);
			if (menuSub.m_hMenu) {
				_DeleteAllMenuItems(menuSub.m_hMenu);
				MTLVERIFY(menuSub.DestroyMenu());
			}
		}

		while (menu.DeleteMenu(0, MF_BYPOSITION))
			;
	}

	void _SaveDirectoryPath(CMenuHandle menu, const CString& strPath)
	{
		m_mapSubMenu.Add(menu.m_hMenu, strPath);
//		m_mapSubMenu[menu.m_hMenu] = strPath;
	}

	CString _LoadDirectoryPath(CMenuHandle menu)
	{
//		return m_mapSubMenu[menu.m_hMenu];
		return m_mapSubMenu.Lookup(menu.m_hMenu);
	}

	int _FindInsertPoint(CMenuHandle menu)
	{
		int nItems = menu.GetMenuItemCount();
//		epmTRACE(_T("_FindInsertPoint ItemCount=%d : %s\n"), nItems, m_strTestText);

		int nInsertPoint;
		for (nInsertPoint = 0; nInsertPoint < nItems; ++nInsertPoint) {
			CMenuItemInfo mii;
			mii.fMask = MIIM_ID;
			menu.GetMenuItemInfo(nInsertPoint, TRUE, &mii);
			if (mii.wID == m_nInsertPointMenuItemID)
				break;
		}

		if (nInsertPoint >= nItems)// not found
			return -1;

		ATLASSERT(nInsertPoint < nItems && "You need a menu item with an ID = m_nInsertPointMenuItemID");
		return nInsertPoint;
	}

	void _DeleteInsertPointMenuItem()
	{
//		epmTRACE(_T("_DeleteInsertPointMenuItem(%d) : %s\n"), m_nInsertPointMenuItemID, m_strTestText);
		m_menu.DeleteMenu(m_nInsertPointMenuItemID, MF_BYCOMMAND);
	}

	CString _GetMenuItemText(const CString& strPath)
	{
		T* pT = static_cast<T*>(this);
		return pT->OnGetMenuItemText(strPath);
	}

	void _RestoreOriginalMenu()
	{
		epmTRACE(_T("_RestoreOriginalMenu\n"));

		while (m_menu.GetMenuItemCount() > m_nOriginalMenuItemCountExceptInsertPointMenuItem) {
			CMenuHandle menuSub = m_menu.GetSubMenu(m_nOriginalInsertPoint);
			if (menuSub.m_hMenu) {
				epmTRACE(_T(" DestroyMenu\n"));
				_DeleteAllMenuItems(menuSub.m_hMenu);
				MTLVERIFY(menuSub.DestroyMenu());
			}
			m_menu.DeleteMenu(m_nOriginalInsertPoint, MF_BYPOSITION);
		}

		m_menu.InsertMenu(m_nOriginalInsertPoint, MF_BYPOSITION, m_nInsertPointMenuItemID, _T("Insert Point"));

//		epmTRACE(_T("_RestoreOriginalMenu(%s) : restored to ItemCount = %d\n"), m_strTestText, m_menu.GetMenuItemCount());
	}

	CString _GetFilePath(const CMenuItemInfo& mii)
	{
		CString str = m_mapID.Lookup(mii.wID);
		if (str.IsEmpty()) {// dir path
			str = m_mapSubMenu.Lookup(mii.hSubMenu);
		}

		ATLASSERT(!str.IsEmpty());
		return str;
	}
};


class CExplorerMenu : public CExplorerMenuImpl<CExplorerMenu>
{
public:
	void OnExecute(const CString& strFilePath)
	{
		::ShellExecute(NULL, _T("open"), strFilePath, NULL, NULL, SW_SHOWNORMAL);
	}
};
/*
template <class T>
class CExplorerMenuPropertyPage : public CPropertyPageImpl<CExplorerMenuPropertyPage>,
	public CWinDataExchange<CExplorerMenuPropertyPage>
{
public:
// Constants
	enum { IDD = IDD_PROPPAGE_EXPMENU };

// Data members
	int m_nAdditionalCheck;
	int m_nMaxMenuItemTextLength;
	int m_nMaxMenuBreakCount;
	T* m_pT;
	HWND m_hWndLinkBar;

// DDX map
	BEGIN_DDX_MAP(CExplorerMenuPropertyPage)
		DDX_CHECK(IDC_CHECK_EXPMENU_ADDITIONAL, m_nAdditionalCheck)
		DDX_INT_RANGE(IDC_EDIT_EXPMENU_ITEMTEXTMAX, m_nMaxMenuItemTextLength, 1, 255)
		DDX_INT_RANGE(IDC_EDIT_EXPMENU_BREAKCOUNTMAX, m_nMaxMenuBreakCount, 2, 5000)
	END_DDX_MAP()

// Constructor
	CExplorerMenuPropertyPage(T* pT, HWND hWndToRefresh)
		: m_pT(pT)
	{
		m_nAdditionalCheck = (m_pT->GetStyle() & EMS_ADDITIONALMENUITEM) ? 1 : 0;
		m_nMaxMenuItemTextLength = m_pT->GetMaxMenuItemTextLength();
		m_nMaxMenuBreakCount = m_pT->GetMaxMenuBreakCount();
		m_hWndLinkBar = hWndToRefresh;
	}

// Overrides
	BOOL OnSetActive()
	{
		SetModified(TRUE);
		return DoDataExchange(FALSE);
	}

	BOOL OnKillActive()
	{
		return DoDataExchange(TRUE);
	}

	BOOL OnApply()
	{
		if (DoDataExchange(TRUE)) {
			DWORD dwStyle = 0;
			if (m_nAdditionalCheck == 1)
				dwStyle |= EMS_ADDITIONALMENUITEM;

			m_pT->SetStyle(dwStyle);
			m_pT->SetMaxMenuItemTextLength(m_nMaxMenuItemTextLength);
			m_pT->SetMaxMenuBreakCount(m_nMaxMenuBreakCount);
			::PostMessage(m_hWndLinkBar, WM_COMMAND, ID_VIEW_REFRESH_LINKBAR, 0);
			return TRUE;
		}
		else
			return FALSE;
	}

// Message map and handlers
	BEGIN_MSG_MAP(CExplorerMenuPropertyPage)
		CHAIN_MSG_MAP(CPropertyPageImpl<CExplorerMenuPropertyPage>)
	END_MSG_MAP()
};


template <class _Profile, class _ExplorerMenu>
void MtlGetProfileExpMenu(_Profile& __profile, _ExplorerMenu& __expmenu)
{
	DWORD dwStyle, dwLength, dwCount;
	LONG lRet = __profile.QueryValue(dwStyle, _T("Style"));
	if (lRet != ERROR_SUCCESS)
		dwStyle = 0;

	lRet = __profile.QueryValue(dwLength,  _T("Max_Text_Length"));
	if (lRet != ERROR_SUCCESS)
		dwLength = 55;

	lRet = __profile.QueryValue(dwCount,  _T("Max_Break_Count"));
	if (lRet != ERROR_SUCCESS)
		dwCount = 5000;

	__expmenu.SetStyle(dwStyle);
	__expmenu.SetMaxMenuItemTextLength(dwLength);
	__expmenu.SetMaxMenuBreakCount(dwCount);
}

template <class _Profile, class _ExplorerMenu>
void MtlWriteProfileExpMenu(_Profile& __profile, _ExplorerMenu& __expmenu)
{
	__profile.SetValue(__expmenu.GetStyle(), _T("Style"));
	__profile.SetValue(__expmenu.GetMaxMenuItemTextLength(), _T("Max_Text_Length"));
	__profile.SetValue(__expmenu.GetMaxMenuBreakCount(), _T("Max_Break_Count"));
}
*/
