#ifndef __MTLMISC_H__
#define __MTLMISC_H__

// MTL Version 0.03
// Copyright (C) 2000 MB<mb2@geocities.co.jp>
// All rights unreserved.
//
// This file is a part of Mb Template Library.
// The code and information is *NOT* provided "as-is" without
// warranty of any kind, either expressed or implied.
//
// Last updated: August 16, 2000

#ifndef __ATLCTRLS_H__
	#error mtlmisc.h requires atlctrls.h to be included first
#endif

#ifndef _SHLOBJ_H_
	#error mtlmisc.h requires shlobj.h to be included first
#endif

#pragma once

#include <algorithm>

namespace MTL
{
/////////////////////////////////////////////////////////////////////////////
// Tips
inline bool MtlIsCrossRect(const CRect& rc1, const CRect& rc2)
{
	CRect rcCross = rc1 & rc2;
	return rcCross.IsRectEmpty() ? false : true;
}

inline int MtlComputeWidthOfText(const CString& strText, HFONT hFont)
{
	if (strText.IsEmpty())
		return 0;

	CWindowDC dc(NULL);
	CRect rcText(0, 0, 0, 0);
	HFONT hOldFont = dc.SelectFont(hFont);
	dc.DrawText(strText, -1, &rcText, DT_SINGLELINE | DT_LEFT | DT_VCENTER | DT_CALCRECT);
	dc.SelectFont(hOldFont);

	return rcText.Width();
}
/////////////////////////////////////////////////////////////////////////////
// Algorithm
inline bool __MtlIsLastChar(const CString& str, TCHAR ch)
{
	if (str.GetLength() <= 0)
		return false;

#ifdef _UNICODE
	if (str[str.GetLength() - 1] == ch)
		return true;
	else
		return false;
#else
	if (str[str.GetLength() - 1] == ch && !_IsDBCSTrailByte(str, str.GetLength() - 1))
		return true;
	else
		return false;
#endif
}

inline TCHAR* Mtl_strstr (
		const TCHAR * str1,
		const TCHAR * str2
		)
{
		TCHAR *cp = (TCHAR *) str1;
		TCHAR *s1, *s2;

		if ( !*str2 )
			return((TCHAR *)str1);

		while (*cp)
		{
				s1 = cp;
				s2 = (TCHAR *) str2;

				/*
				 * MBCS: ok to ++ since doing equality comparison.
				 * [This depends on MBCS strings being "legal".]
				 */
				while ( *s1 && *s2 && (*s1 == *s2) )
					s1++, s2++;

				if (!*s2)
						return(cp);

				cp = ::CharNext(cp);
		}

		return(NULL);
}

// CString::Find is too late...
inline int __MtlFind(LPCTSTR lpszSrc, LPCTSTR lpszSub)
{
	// find first matching substring
	LPTSTR lpsz = Mtl_strstr(lpszSrc, lpszSub);

	// return -1 for not found, distance from beginning otherwise
	return (lpsz == NULL) ? -1 : (int)(lpsz - lpszSrc);
}

inline int __MtlFind(const CString& str, int nStart, const CString& strFind)
{
	CString strText = str.Right(str.GetLength() - nStart);
	int nIndex = strText.Find(strFind);//__MtlFind(strText, strFind);
	if (nIndex == -1)
		return -1;
	else {
		return nStart + nIndex;
	}
}

inline int __MtlFindChar(const CString& str, int nStart, TCHAR ch)
{
	CString strText = str.Right(str.GetLength() - nStart);
	int nIndex = strText.Find(ch);
	if (nIndex == -1)
		return -1;
	else {
		ATLASSERT(str[nStart + nIndex] == ch);
		return nStart + nIndex;
	}
}

template <class _InputIter>
bool MtlSearchString(_InputIter __first, _InputIter __last, const CString& str)
{
	if (str.IsEmpty())
		return false;

	for ( ; __first != __last; ++__first) {
		if (str.Find(*__first) != -1)
			return true;
	}

	return false;
}

inline TCHAR Mtl_tolower(TCHAR ch)
{
	return ((ch >= _T('A') && ch <= _T('Z')) ? (ch - _T('A') + _T('a')) : ch);
}

inline bool Mtl_istalnum(TCHAR ch)
{
	return (ch >= _T('a') && ch <= _T('z')) ||
		(ch >= _T('A') && ch <= _T('Z')) || (ch >= _T('0') && ch <= _T('9'));
}

inline bool Mtl_isalpha(TCHAR ch)
{
	return (ch >= _T('a') && ch <= _T('z')) || (ch >= _T('A') && ch <= _T('Z'));
}

//Thanks to http://www.hidecnet.ne.jp/~sinzan/tips/main.htm, and you.
inline bool _StrMatchWildCard(LPCTSTR Ptn, LPCTSTR Str)
{
	switch( *Ptn )
	{
		case _T('\0'):
			return (*Str==_T('\0'));
		case _T('*'):
			return _StrMatchWildCard( Ptn+1, Str ) || (*Str!=_T('\0')) && _StrMatchWildCard( Ptn, Str+1 );
		case _T('?'):
			return (*Str!=_T('\0')) && _StrMatchWildCard( Ptn+1, Str+1 );
		default:
			return (Mtl_tolower(*Ptn) == Mtl_tolower(*Str)) && _StrMatchWildCard(Ptn + 1, Str + 1);
	}
}

template <class _InputIter>
bool MtlSearchStringWildCard(_InputIter __first, _InputIter __last, const CString& str)
{
	if (str.IsEmpty())
		return false;

	for ( ; __first != __last; ++__first) {
		CString ptn(*__first);
		if (ptn.Find(_T('*')) != -1 || ptn.Find(_T('?')) != -1) {
			if (_StrMatchWildCard(ptn, str))
				return true;
		}
		else {
			if (str == *__first) {
				ATLTRACE2(atlTraceUser, 4, _T("%s = %s\n"), *__first, str);
				return true;
			}
		}
	}

	return false;
}

inline void MtlRemoveTrailingBackSlash(CString& str)
{
	if (__MtlIsLastChar(str, _T('\\')))
		str = str.Left(str.GetLength() - 1);
/*
	int nIndex = str.ReverseFind(_T('\\'));
	if (nIndex != -1 && nIndex == str.GetLength() - 1)// trailing
		str = str.Left(str.GetLength() - 1);
*/
}

inline void MtlRemoveStringHeaderAndFooter(CString& str)
{
	if (str.GetLength() == 0)
		return;

	if (str.GetAt(0) == _T('\"') && __MtlIsLastChar(str, _T('\"')))
		str = str.Mid(1, str.GetLength() - 2);
}

inline void MtlRemoveTrailingSpace(CString& str)
{
	if (__MtlIsLastChar(str, _T(' ')))
		str = str.Left(str.GetLength() - 1);
}

inline void MtlMakeSureTrailingChar(CString& strDirectoryPath, TCHAR ch)
{
	if (!__MtlIsLastChar(strDirectoryPath, ch))
		strDirectoryPath += ch;
}

inline void MtlMakeSureTrailingBackSlash(CString& strDirectoryPath)
{
	MtlMakeSureTrailingChar(strDirectoryPath, _T('\\'));
}

inline CString MtlGetDirectoryPath(const CString& strPath, bool bAddBackSlash = false, TCHAR chSep = _T('\\'))
{
	int nIndex = strPath.ReverseFind(chSep);
	CString strRet = strPath.Left(nIndex);

	if (bAddBackSlash)
		MtlMakeSureTrailingChar(strRet, chSep);

	return strRet;
}

template <class _Function>
bool MtlForEachObject_OldShell(const CString& strDirectoryPath, _Function __f, bool bIncludeHidden = false)
{
	CString strPathFind = strDirectoryPath;
	MtlMakeSureTrailingBackSlash(strPathFind);
	CString strPath = strPathFind;
	strPathFind += _T("*.*");

	WIN32_FIND_DATA wfd;
	HANDLE h = ::FindFirstFile(strPathFind, &wfd);
	if (h == INVALID_HANDLE_VALUE)
		return false;

	// Now scan the directory
	do {
		BOOL bDirectory = wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY;
		BOOL bVisible = bIncludeHidden ? TRUE : ((wfd.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) == 0);

		if (bVisible == FALSE)
			continue;

		if (bDirectory) {
			// ignore the current and parent directory entries
			if (::lstrcmp(wfd.cFileName, _T(".")) == 0 || ::lstrcmp(wfd.cFileName, _T("..")) == 0)
				continue;

			CString strDirectoryPath = strPath + wfd.cFileName;
			MtlMakeSureTrailingBackSlash(strDirectoryPath);
			__f(strDirectoryPath, true);
			continue;
		}
		else {
			__f(strPath + wfd.cFileName, false);
		}

	} while (::FindNextFile(h, &wfd));
	::FindClose(h);

	return true;
}

template <class _Function>
bool MtlForEachFile(const CString& strDirectoryPath, _Function __f)
{
	CString strPathFind = strDirectoryPath;
	MtlMakeSureTrailingBackSlash(strPathFind);
	CString strPath = strPathFind;
	strPathFind += _T("*.*");

	WIN32_FIND_DATA wfd;
	HANDLE h = ::FindFirstFile(strPathFind, &wfd);
	if (h == INVALID_HANDLE_VALUE)
		return false;

	// Now scan the directory
	do {
		// it is a file
		if ((wfd.dwFileAttributes & (FILE_ATTRIBUTE_DIRECTORY|FILE_ATTRIBUTE_HIDDEN|FILE_ATTRIBUTE_SYSTEM)) == 0) {
			__f(strPath + wfd.cFileName);
		}
	} while (::FindNextFile(h, &wfd));
	::FindClose(h);

	return true;
}

inline bool MtlIsFileExist(const CString& strDirectoryPath)
{
	CString strPathFind = strDirectoryPath;
	MtlMakeSureTrailingBackSlash(strPathFind);
	CString strPath = strPathFind;
	strPathFind += _T("*.*");

	WIN32_FIND_DATA wfd;
	HANDLE h = ::FindFirstFile(strPathFind, &wfd);
	if (h == INVALID_HANDLE_VALUE)
		return false;

	// Now scan the directory
	do {
		// it is a file
		if ((wfd.dwFileAttributes & (FILE_ATTRIBUTE_DIRECTORY|FILE_ATTRIBUTE_HIDDEN|FILE_ATTRIBUTE_SYSTEM)) == 0) {
			::FindClose(h);
			return true;
		}
	} while (::FindNextFile(h, &wfd));

	::FindClose(h);
	return false;
}

template <class _Function>
bool MtlForEachFileSort(const CString& strDirectoryPath, _Function __f)
{
	CString strPathFind = strDirectoryPath;
	CSimpleArray<CString> strings;
	MtlMakeSureTrailingBackSlash(strPathFind);
	CString strPath = strPathFind;
	strPathFind += _T("*.*");

	WIN32_FIND_DATA wfd;
	HANDLE h = ::FindFirstFile(strPathFind, &wfd);
	if (h == INVALID_HANDLE_VALUE)
		return false;

	// Now scan the directory
	do {
		// it is a file
		if ((wfd.dwFileAttributes & (FILE_ATTRIBUTE_DIRECTORY|FILE_ATTRIBUTE_HIDDEN|FILE_ATTRIBUTE_SYSTEM)) == 0) {
			strings.Add(strPath + wfd.cFileName);
		}
	} while (::FindNextFile(h, &wfd));
	::FindClose(h);

	std::sort(_begin(strings), _end(strings));
	std::for_each(_begin(strings), _end(strings), __f);

	return true;
}

template <class _Function>
bool MtlForEachDirectory(const CString& strDirectoryPath, _Function __f)
{
	CString strPathFind = strDirectoryPath;
	MtlMakeSureTrailingBackSlash(strPathFind);
	CString strPath = strPathFind;
	strPathFind += _T("*.*");

	WIN32_FIND_DATA wfd;
	HANDLE h = ::FindFirstFile(strPathFind, &wfd);
	if (h == INVALID_HANDLE_VALUE)
		return false;

	// check the subdirectories for more
	do {
		BOOL bDirectory = wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY;
		BOOL bVisible = (wfd.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) == 0;
		if (bDirectory && bVisible) {
			// ignore the current and parent directory entries
			if (::lstrcmp(wfd.cFileName, _T(".")) == 0 || ::lstrcmp(wfd.cFileName, _T("..")) == 0)
				continue;

			CString strDirectoryPath = strPath + wfd.cFileName;
			MtlMakeSureTrailingBackSlash(strDirectoryPath);
			__f(strDirectoryPath);
		}
	} while (::FindNextFile(h, &wfd));
	::FindClose(h);

	return true;
}

template <class _Function>
bool MtlForEachDirectorySort(const CString& strDirectoryPath, _Function __f)
{
	CString strPathFind = strDirectoryPath;
	CSimpleArray<CString> strings;
	MtlMakeSureTrailingBackSlash(strPathFind);
	CString strPath = strPathFind;
	strPathFind += _T("*.*");

	WIN32_FIND_DATA wfd;
	HANDLE h = ::FindFirstFile(strPathFind, &wfd);
	if (h == INVALID_HANDLE_VALUE)
		return false;

	// check the subdirectories for more
	do {
		BOOL bDirectory = wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY;
		BOOL bVisible = (wfd.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) == 0;
		if (bDirectory && bVisible) {
			// ignore the current and parent directory entries
			if (::lstrcmp(wfd.cFileName, _T(".")) == 0 || ::lstrcmp(wfd.cFileName, _T("..")) == 0)
				continue;

			CString strDirectoryPath = strPath + wfd.cFileName;
			MtlMakeSureTrailingBackSlash(strDirectoryPath);
			strings.Add(strDirectoryPath);
		}
	} while (::FindNextFile(h, &wfd));
	::FindClose(h);

	std::sort(_begin(strings), _end(strings));
	std::for_each(_begin(strings), _end(strings), __f);
	return true;
}

struct _MtlEnumMDIChildStruct
{
	CSimpleArray<HWND> _arrChildWnd;
	HWND _hWndMDIClient;
};

template <class _Function>
_Function MtlForEachMDIChild(HWND hWndMDIClient, _Function __f)
{
	_MtlEnumMDIChildStruct enumstruct;
	enumstruct._hWndMDIClient = hWndMDIClient;

	::EnumChildWindows(hWndMDIClient, &_MtlMDIChildEnumProc, (LPARAM)&enumstruct);

	return std::for_each(_begin(enumstruct._arrChildWnd), _end(enumstruct._arrChildWnd), __f);
}

static BOOL CALLBACK _MtlMDIChildEnumProc(HWND hWnd, LPARAM lParam)
{
	if (::GetWindow(hWnd, GW_OWNER))
		return TRUE;

	_MtlEnumMDIChildStruct* penumstruct = (_MtlEnumMDIChildStruct*)lParam;
	if (::GetParent(hWnd) != penumstruct->_hWndMDIClient)
		return TRUE;

	penumstruct->_arrChildWnd.Add(hWnd);
	return TRUE;
}

struct _MtlEnumChildren
{
	CSimpleArray<HWND> _arrChildWnd;
};

template <class _Function>
_Function MtlForEachChildren(HWND hWndParent, _Function __f)
{
	_MtlEnumChildren enumstruct;

	::EnumChildWindows(hWndParent, &_MtlChildrenEnumProc, (LPARAM)&enumstruct);

	return std::for_each(_begin(enumstruct._arrChildWnd), _end(enumstruct._arrChildWnd), __f);
}

static BOOL CALLBACK _MtlChildrenEnumProc(HWND hWnd, LPARAM lParam)
{
	_MtlEnumChildren* penumstruct = (_MtlEnumChildren*)lParam;
	penumstruct->_arrChildWnd.Add(hWnd);
	return TRUE;
}

struct _MtlEnumChildrenCount
{
	HWND _hWndParent;
	int _nCount;
};

static BOOL CALLBACK _MtlChildrenCountEnumProc(HWND hWnd, LPARAM lParam)
{
	_MtlEnumChildrenCount* penumstruct = (_MtlEnumChildrenCount*)lParam;
	if (::GetParent(hWnd) == penumstruct->_hWndParent)
		++penumstruct->_nCount;

	return TRUE;
}

inline int MtlGetDirectChildrenCount(HWND hWndParent)
{
	_MtlEnumChildrenCount enumstruct = { hWndParent, 0 };

	::EnumChildWindows(hWndParent, &_MtlChildrenCountEnumProc, (LPARAM)&enumstruct);

	return enumstruct._nCount;
}

//////////////////////////////////////////////////////////////////////////
struct _MtlEnumChildrenName
{
	HWND _hWndParent;
	LPCTSTR _lpszClassName;
	int _nCount;
};

static BOOL CALLBACK _MtlIsDirectChildrenNameEnumProc(HWND hWnd, LPARAM lParam)
{
	_MtlEnumChildrenName* penumstruct = (_MtlEnumChildrenName*)lParam;
	if (::GetParent(hWnd) == penumstruct->_hWndParent) {
		TCHAR sz[50];
		::GetClassName(hWnd, sz, 50);
		if (::lstrcmp(sz, penumstruct->_lpszClassName) == 0) {
			penumstruct->_nCount++;
		}
	}

	return TRUE;
}

inline int MtlCountDirectChildrenName(HWND hWndParent, const CString& strClassName)
{
	_MtlEnumChildrenName enumstruct = { hWndParent, (LPCTSTR)strClassName, 0 };

	::EnumChildWindows(hWndParent, &_MtlIsDirectChildrenNameEnumProc, (LPARAM)&enumstruct);

	return enumstruct._nCount;
}
//////////////////////////////////////////////////////////////////////////
inline CString MtlGetClipboardText(bool bUseOLE = false)
{
	CString strText;

	if (!bUseOLE) {
		if (::IsClipboardFormatAvailable(CF_TEXT) && ::OpenClipboard(NULL)) {
			HGLOBAL hText = ::GetClipboardData(CF_TEXT);
			if (hText) {
				strText = reinterpret_cast<LPSTR>(::GlobalLock(hText));
				::GlobalUnlock(hText);
			}
			::CloseClipboard();
		}
	}
	else {
		CComPtr<IDataObject> spDataObject;
		HRESULT hr = ::OleGetClipboard(&spDataObject);
		if (FAILED(hr))
			return strText;

		FORMATETC formatetc = { CF_TEXT, NULL, DVASPECT_CONTENT, -1, TYMED_HGLOBAL };
		STGMEDIUM stgmedium;
		hr = spDataObject->GetData(&formatetc, &stgmedium);
		if (SUCCEEDED(hr)) {
			if (stgmedium.hGlobal != NULL) {
				HGLOBAL hText = stgmedium.hGlobal;
				strText = reinterpret_cast<LPSTR>(::GlobalLock(hText));
				::GlobalUnlock(hText);
			}
			::ReleaseStgMedium(&stgmedium);
		}
	}

	return strText;
}

inline bool MtlSetClipboardText(const CString& str, HWND hWnd)
{
	if (str.IsEmpty())
		return false;

	int nByte = str.GetLength();
	HGLOBAL hText = ::GlobalAlloc(GMEM_MOVEABLE | GMEM_ZEROINIT, nByte + 1);
	if (hText == NULL)
		return false;

	BYTE* pText = (BYTE*)::GlobalLock(hText);
	if (pText == NULL)
		return false;

	::memcpy(pText, (LPCTSTR)str, nByte);

	::GlobalUnlock(hText);

	::OpenClipboard(hWnd);
	::EmptyClipboard();
	::SetClipboardData(CF_TEXT, hText);
	::CloseClipboard();

	return true;
}

inline bool AtlCompactPathFixed(LPTSTR lpstrOut, LPCTSTR lpstrIn, int cchLen)
{
	ATLASSERT(lpstrOut != NULL);
	ATLASSERT(lpstrIn != NULL);
	ATLASSERT(cchLen > 0);

	LPCTSTR szEllipsis = _T("...");
	const int cchEndEllipsis = 3;
	const int cchMidEllipsis = 4;

	if(lstrlen(lpstrIn) < cchLen)	// ******changed****** I can't understand why (+1) needed.
		return (lstrcpy(lpstrOut, lpstrIn) != NULL);

	*lpstrOut = _T('\0');	// ******added******
							// As lstrcat needs NULL-terminated string, lstrcat can't touch lpstrOut without this,
							// and lpstrOut will be endless string and it may crash your program.

	// check if the separator is a slash or a backslash
	TCHAR chSlash = _T('\\');
	for(LPTSTR lpstr = (LPTSTR)lpstrIn; *lpstr != 0; lpstr = ::CharNext(lpstr))
	{
		if((*lpstr == _T('/')) || (*lpstr == _T('\\')))
			chSlash = *lpstr;
	}

	// find the filename portion of the path
	LPCTSTR lpstrFileName = lpstrIn;
	for(LPCTSTR pPath = lpstrIn; *pPath; pPath = ::CharNext(pPath))
	{
		if((pPath[0] == _T('\\') || pPath[0] == _T(':') || pPath[0] == _T('/'))
				&& pPath[1] && pPath[1] != _T('\\') && pPath[1] != _T('/'))
			lpstrFileName = pPath + 1;
	}
	int cchFileName = lstrlen(lpstrFileName);

	// handle just the filename without a path
	if(lpstrFileName == lpstrIn && cchLen > cchEndEllipsis)
	{
		bool bRet = (lstrcpyn(lpstrOut, lpstrIn, cchLen - cchEndEllipsis) != NULL);
		if(bRet)
		{
#ifndef _UNICODE
			// The index that must be checked was wrong.
			if(_IsDBCSTrailByte(lpstrIn, cchLen - cchEndEllipsis - 1)) {	// ******changed******
				ATLASSERT(cchLen - cchEndEllipsis - 2 >= 0);				// ******added******
				lpstrOut[cchLen - cchEndEllipsis - 2] = 0;					// ******changed******
			}
#endif //_UNICODE
			bRet = (lstrcat(lpstrOut, szEllipsis) != NULL);
		}
		return bRet;
	}

	// handle just ellipsis
	if((cchLen < (cchMidEllipsis + cchEndEllipsis)))
	{
		int i;	//+++
		for(/*int*/ i = 0; i < cchLen - 1; i++)
			lpstrOut[i] = ((i + 1) == cchMidEllipsis) ? chSlash : _T('.');
		lpstrOut[i] = 0;
		return true;
	}

	// calc how much we have to copy
	int cchToCopy = cchLen - (cchMidEllipsis + cchFileName);

	if(cchToCopy < 0)
		cchToCopy = 0;

#ifndef _UNICODE
	if(cchToCopy > 0 && _IsDBCSTrailByte(lpstrIn, cchToCopy - 1)) // ******changed******
		cchToCopy--;
#endif //_UNICODE

	bool bRet = (lstrcpyn(lpstrOut, lpstrIn, cchToCopy) != NULL);
	if(!bRet)
		return false;

	// add ellipsis
	bRet = (lstrcat(lpstrOut, szEllipsis) != NULL);
	if(!bRet)
		return false;
	TCHAR szSlash[2] = { chSlash, 0 };
	bRet = (lstrcat(lpstrOut, szSlash) != NULL);
	if(!bRet)
		return false;

	// add filename (and ellipsis, if needed)
	if(cchLen > (cchMidEllipsis + cchFileName))
	{
		bRet = (lstrcat(lpstrOut, lpstrFileName) != NULL);
	}
	else
	{
		cchToCopy = cchLen - cchMidEllipsis - cchEndEllipsis;
#ifndef _UNICODE
		if(cchToCopy > 0 && _IsDBCSTrailByte(lpstrFileName, cchToCopy - 1)) // ******changed******
			cchToCopy--;
#endif //_UNICODE
		bRet = (lstrcpyn(&lpstrOut[cchMidEllipsis], lpstrFileName, cchToCopy) != NULL);
		if(bRet)
			bRet = (lstrcat(lpstrOut, szEllipsis) != NULL);
	}

	return bRet;
}

inline bool MtlCompactPath(CString& strOut, const CString& str, int cchLen)
{
	ATLASSERT(cchLen > 0);

	bool bRet = AtlCompactPathFixed(strOut.GetBufferSetLength(cchLen), str, cchLen);
	strOut.ReleaseBuffer();

	if (!bRet || strOut.IsEmpty() || strOut == _T("..."))
		return false;

	return true;
}

inline CString MtlCompactString(const CString& str, int nMaxTextLength)
{
	ATLASSERT(nMaxTextLength > 0);
	if (str.GetLength() <= nMaxTextLength)
		return str;

	LPCTSTR szEllipsis = _T("...");
	const int cchEndEllipsis = 3;

	if (nMaxTextLength <= cchEndEllipsis) {// only ellipsis
		return CString(_T('.'), nMaxTextLength);
	}

	int nIndex = nMaxTextLength - cchEndEllipsis;
	ATLASSERT(nIndex > 0);
	ATLASSERT(nIndex < str.GetLength());

	if (_IsDBCSTrailByte(str, nIndex))
		--nIndex;// one step back

	ATLASSERT(nIndex >= 0);

	return str.Left(nIndex) + szEllipsis;
}

inline void MtlRefreshBandIdealSize(CReBarCtrl rebar, CToolBarCtrl toolbar)
{
	REBARBANDINFO rbBand;
	rbBand.cbSize = sizeof(REBARBANDINFO);
	rbBand.fMask = RBBIM_IDEALSIZE;

	// Calculate the size of the band
	int nBtnCount = toolbar.GetButtonCount();
	if(nBtnCount > 0)
	{
		RECT rcTmp;
		BOOL bRet = (BOOL)toolbar.GetItemRect(nBtnCount-1, &rcTmp);
		ATLASSERT(bRet);
		rbBand.cxIdeal = rcTmp.right;
		int nIndex = rebar.IdToIndex(toolbar.GetDlgCtrlID());
		rebar.SetBandInfo(nIndex, &rbBand);
	}
}

inline CString MtlCurrentDirectoryFileName(const CString& strFileName)
{
	TCHAR sz[MAX_PATH];
	::GetModuleFileName(_Module.GetModuleInstance(), sz, MAX_PATH);

	CString str(sz);
	int nIndex = str.ReverseFind(_T('\\'));

	return str.Left(nIndex + 1) + strFileName;
}

inline CString MtlGetModuleFileName()
{
	TCHAR sz[MAX_PATH];
	::GetModuleFileName(_Module.GetModuleInstance(), sz, MAX_PATH);
	return sz;
}

inline CString MtlGetFileNameFromCommandLine(const CString& _strCommand)
{
	CString strCommand = _strCommand;
	if (strCommand.Left(1) == _T("\"")) {
		strCommand = strCommand.Right(strCommand.GetLength() - 1);
	}
	else {
		return strCommand;
	}

	int i = strCommand.Find('"');
	if (i == -1)
		return _strCommand; // illegally terminated

	return strCommand.Left(i);
}

inline BOOL MtlSetExcutable(const CString& strExt, const CString& strExe_, const CString& strName, bool bOn)
{
	CString strExe = _T('"') + strExe_ + _T('"');// fixed by Shimawari, thanks!

	CRegKey rkExt;
	CRegKey rkShell;
	CRegKey rkDonut;
	CRegKey rkCommand;

	LONG lRet;
	// open ROOT
	lRet = rkExt.Create(HKEY_CLASSES_ROOT, strExt);
	if(lRet != ERROR_SUCCESS)
		return FALSE;


	if (bOn) {
		// set shell value to name (means default)
		lRet = rkExt.SetKeyValue(_T("shell"), strName);
		if(lRet != ERROR_SUCCESS)
			return FALSE;

		// open shell
		lRet = rkShell.Create(rkExt, _T("shell"));
		if(lRet != ERROR_SUCCESS)
			return FALSE;

		lRet = rkDonut.Create(rkShell, strName);
		if(lRet != ERROR_SUCCESS)
			return FALSE;
		lRet = rkDonut.SetValue(strName);
		ATLASSERT(lRet == ERROR_SUCCESS);

		// command has /dde
		lRet = rkCommand.Create(rkDonut, _T("command"));
		if(lRet != ERROR_SUCCESS)
			return FALSE;
		lRet = rkCommand.SetValue(strExe + _T(" /dde"));//\"%1\""));
		ATLASSERT(lRet == ERROR_SUCCESS);

		// DDE
		CRegKey rkDDEExec;		// order is important
		CRegKey rkApplication;
		CRegKey rkTopic;

		lRet = rkDDEExec.Create(rkDonut, _T("ddeexec"));
		if(lRet != ERROR_SUCCESS)
			return FALSE;
		lRet = rkDDEExec.SetValue(_T("[open(\"%1\")]"));
		ATLASSERT(lRet == ERROR_SUCCESS);

		lRet = rkApplication.Create(rkDDEExec, _T("application"));
		if(lRet != ERROR_SUCCESS)
			return FALSE;
		lRet = rkApplication.SetValue(strName);
		ATLASSERT(lRet == ERROR_SUCCESS);

		lRet = rkTopic.Create(rkDDEExec, _T("topic"));
		if(lRet != ERROR_SUCCESS)
			return FALSE;
		lRet = rkTopic.SetValue(_T("system"));
		ATLASSERT(lRet == ERROR_SUCCESS);
	}
	else {
		// open shell
		lRet = rkShell.Open(rkExt, _T("shell"));
		if(lRet != ERROR_SUCCESS)
			return FALSE;

		// clean up name if my name
		DWORD nLen;
		lRet = rkShell.QueryValue(NULL, NULL, &nLen);
		if(lRet != ERROR_SUCCESS)
			return FALSE;
		ATLTRACE2(atlTraceUser, 4, _T("MtlIsExecutable :%d\n"), nLen);

		LPTSTR lpszText = (LPTSTR)_alloca((nLen + 1) * sizeof(TCHAR));
		lRet = rkShell.QueryValue(lpszText, NULL, &nLen);
		if (::lstrcmp(lpszText, strName) == 0) {
			lRet = rkExt.SetKeyValue(_T("shell"), _T(""));
			if(lRet != ERROR_SUCCESS)
				return FALSE;
		}

		lRet = rkShell.DeleteSubKey(strName);
		if(lRet != ERROR_SUCCESS)
			return FALSE;
	}

	return TRUE;
}

inline BOOL MtlIsExecutable(const CString& strExt, const CString& strName)
{
	CRegKey rkExt;
	CRegKey rkShell;

	LONG lRet;
	// open ROOT
	lRet = rkExt.Open(HKEY_CLASSES_ROOT, strExt);
	if(lRet != ERROR_SUCCESS)
		return FALSE;

	// open shell
	lRet = rkShell.Open(rkExt, _T("shell"));
	if(lRet != ERROR_SUCCESS)
		return FALSE;

	DWORD nLen;
	lRet = rkShell.QueryValue(NULL, NULL, &nLen);
	if(lRet != ERROR_SUCCESS)
		return FALSE;
	ATLTRACE2(atlTraceUser, 4, _T("MtlIsExecutable :%d\n"), nLen);

	LPTSTR lpszText = (LPTSTR)_alloca((nLen + 1) * sizeof(TCHAR));
	lRet = rkShell.QueryValue(lpszText, NULL, &nLen);
	ATLTRACE2(atlTraceUser, 4, _T("MtlIsExecutable :%s\n"), lpszText);
	if (::lstrcmp(lpszText, strName) == 0)
		return TRUE;
	else
		return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
// MtlOpenHttpFiles
inline bool MtlIsInternetFileMaybe(const CString& strPath)
{
	if (strPath.Right(1) == _T("/"))// all I can do
		return true;

	int nIndex = strPath.ReverseFind(_T('/'));
	if (nIndex == -1)
		return false;

	nIndex = __MtlFindChar(strPath, nIndex, _T('.'));
	if (nIndex == -1)
		return false;

	CString strExt3 = strPath.Mid(nIndex, 4);
	CString strExt4 = strPath.Mid(nIndex, 5);
	CString strExt5 = strPath.Mid(nIndex, 6);

	if (strExt4.CompareNoCase(_T(".html")) == 0 ||
		strExt5.CompareNoCase(_T(".shtml")) == 0 ||
		strExt3.CompareNoCase(_T(".htm")) == 0 ||
		strExt3.CompareNoCase(_T(".url")) == 0 ||
		strExt3.CompareNoCase(_T(".cgi")) == 0 ||
		strExt3.CompareNoCase(_T(".asp")) == 0)
		return true;
	else
		return false;
}

bool __MtlIsMeaninglessCharInHttp(TCHAR ch)
{
	return ch == _T('\\') || ch == _T(':') ||
		ch == _T(',') || ch == _T(';') ||
		ch == _T('*') || ch == _T('\"') ||
		ch == _T('<') || ch == _T('>') ||
		ch == _T('|') ||
		ch == _T(' ') || ch == _T('\t') ||
		ch == _T('\r');
}

inline CString __MtlRemoveTrailingMeaninglessChar(const CString& str_)
{
	ATLTRACE2(atlTraceUser, 4, _T("__MtlRemoveTrailingMeaninglessChar: (%s)\n"), str_);
	CString str = str_;
	int nCount = 0;
	for (int i = str.GetLength() - 1; i >= 0; --i) {
		ATLTRACE2(atlTraceUser, 4, _T("                                  : (%d, %d)\n"), str[i], i);
		if (__MtlIsMeaninglessCharInHttp(str[i])) {
			nCount++;
		}
		else
			break;
	}

	return str.Left(str.GetLength() - nCount);
}

void MtlOpenHttpFiles(CSimpleArray<CString>& arrFiles, const CString& strCmdLine_)
{
	ATLTRACE2(atlTraceUser, 4, _T("MtlOpenHttpFiles :%s\n"), strCmdLine_.Left(100));
	CString strCmdLine = strCmdLine_;

	if (strCmdLine.IsEmpty())
		return;

	int nFirst = 0;
	int nLast;
	CString strFile;
	do {
		ATLTRACE2(atlTraceUser, 4, _T(" step1\n"));

		nFirst = __MtlFind(strCmdLine, nFirst, _T("http:"));
		if (nFirst == -1)
			break;

		ATLTRACE2(atlTraceUser, 4, _T(" step2\n"));

		nLast = __MtlFindChar(strCmdLine, nFirst + 4, _T('\n'));
		if (nLast == -1) {
			strFile = strCmdLine.Mid(nFirst);
		}
		else {
			ATLASSERT(strCmdLine.GetAt(nLast) == _T('\n'));
			strFile = strCmdLine.Mid(nFirst, nLast - nFirst);
			ATLASSERT(strFile.Right(1) != _T('\n'));
		}

		ATLTRACE2(atlTraceUser, 4, _T(" step3\n"));
		strFile = __MtlRemoveTrailingMeaninglessChar(strFile);
		ATLTRACE2(atlTraceUser, 4, _T("                 :cleaned(%s)\n"), strFile);
		if (MtlIsInternetFileMaybe(strFile)) {
			ATLTRACE2(atlTraceUser, 4, _T("                 :open   (%s)\n"), strFile);
			arrFiles.Add(strFile);
		}

		if (nLast == -1)
			break;

		ATLTRACE2(atlTraceUser, 4, _T(" step4\n"));
		nFirst = nLast;
	} while (nFirst < strCmdLine.GetLength());
}

template <class T>
class CDDEMessageHandler
{
public:
// Data members
	ATOM m_atomApp, m_atomSystemTopic;

// Constructor/Destructor
	CDDEMessageHandler(const CString& strAppName)
	{
		m_atomApp = ::GlobalAddAtom(strAppName);
		m_atomSystemTopic = ::GlobalAddAtom(_T("system"));
	}

	~CDDEMessageHandler()
	{
		::GlobalDeleteAtom(m_atomApp);
		::GlobalDeleteAtom(m_atomSystemTopic);

#ifdef _DEBUG
//		TCHAR szAtomName[_MAX_PATH];
//		ATLASSERT(::GlobalGetAtomName(m_atomApp, szAtomName, _MAX_PATH - 1) == 0);
//		ATLASSERT(::GlobalGetAtomName(m_atomSystemTopic, szAtomName, _MAX_PATH - 1) == 0);
#endif
	}

// Overridables
	bool OnDDEOpenFile(const CString& strFileName)
	{
		return true;
	}

// Message map and handlers
	BEGIN_MSG_MAP(CDDEMessageHandler)
		MESSAGE_HANDLER(WM_DDE_INITIATE, OnDDEInitiate)
		MESSAGE_HANDLER(WM_DDE_EXECUTE, OnDDEExecute)
		MESSAGE_HANDLER(WM_DDE_TERMINATE, OnDDETerminate)
	END_MSG_MAP()

	LRESULT OnDDEInitiate(UINT /*uMsg*/, WPARAM wParam, LPARAM lParam, BOOL& /*bHandled*/)
	{
		if (LOWORD(lParam) != 0 && HIWORD(lParam) != 0 &&
			(ATOM)LOWORD(lParam) == m_atomApp &&
			(ATOM)HIWORD(lParam) == m_atomSystemTopic)
		{
			ATLTRACE2(atlTraceUser, 4, _T("OnDDEInitiate - from windows shell\n"));
			// make duplicates of the incoming atoms (really adding a reference)
			TCHAR szAtomName[_MAX_PATH];
			MTLVERIFY(::GlobalGetAtomName(m_atomApp,
				szAtomName, _MAX_PATH - 1) != 0);
			MTLVERIFY(::GlobalAddAtom(szAtomName) == m_atomApp);
			MTLVERIFY(::GlobalGetAtomName(m_atomSystemTopic,
				szAtomName, _MAX_PATH - 1) != 0);
			MTLVERIFY(::GlobalAddAtom(szAtomName) == m_atomSystemTopic);

			// send the WM_DDE_ACK (caller will delete duplicate atoms)
			T* pT = static_cast<T*>(this);
			::SendMessage((HWND)wParam, WM_DDE_ACK, (WPARAM)pT->m_hWnd,
				MAKELPARAM(m_atomApp, m_atomSystemTopic));
		}
		return 0L;
	}

// always ACK the execute command - even if we do nothing

	LRESULT OnDDEExecute(UINT /*uMsg*/, WPARAM wParam, LPARAM lParam, BOOL& /*bHandled*/)
	{
		ATLTRACE2(atlTraceUser, 4, _T("OnDDEExecute - from windows shell\n"));
		// unpack the DDE message
		UINT unused;
		HGLOBAL hData;
		MTLVERIFY(UnpackDDElParam(WM_DDE_EXECUTE, lParam, &unused, (UINT*)&hData));

		// get the command string
		CString strCommand(reinterpret_cast<LPCTSTR>( ::GlobalLock(hData) ));
		GlobalUnlock(hData);

		T* pT = static_cast<T*>(this);
		// acknowledge now - before attempting to execute
		::PostMessage((HWND)wParam, WM_DDE_ACK, (WPARAM)pT->m_hWnd,
			ReuseDDElParam(lParam, WM_DDE_EXECUTE, WM_DDE_ACK,
			(UINT)0x8000, (UINT)hData));

		// don't execute the command when the window is disabled
		if (!pT->IsWindowEnabled())
		{
			ATLTRACE(_T("Warning: DDE command '%s' ignored because window is disabled.\n"),
				strCommand);
			return 0;
		}

		// execute the command
		if (!_OnDDECommand(strCommand))
			ATLTRACE(_T("Error: failed to execute DDE command '%s'.\n"), strCommand);

		return 0L;
	}

	LRESULT OnDDETerminate(UINT /*uMsg*/, WPARAM wParam, LPARAM lParam, BOOL& /*bHandled*/)
	{
		T* pT = static_cast<T*>(this);
		::PostMessage((HWND)wParam, WM_DDE_TERMINATE, (WPARAM)pT->m_hWnd, lParam);
		return 0L;
	}

	bool _OnDDECommand(const CString& _strCommand)
	{
		CString strCommand = _strCommand;

		// open format is "[open("%s")]" - no whitespace allowed, one per line
		if (strCommand.Left(7) == _T("[open(\""))
		{
			strCommand = strCommand.Right(strCommand.GetLength() - 7);
		}
		else {
			return false;
		}

		int i = strCommand.Find('"');
		if (i == -1)
			return false; // illegally terminated

		CString strFileName = strCommand.Left(i);

		T* pT = static_cast<T*>(this);
		return pT->OnDDEOpenFile(strFileName);
	}
};

void MtlDrawDragRectFixed(CDCHandle dc, LPCRECT lpRect, SIZE size, LPCRECT lpRectLast, SIZE sizeLast, HBRUSH hBrush = NULL, HBRUSH hBrushLast = NULL)
{
	// first, determine the update region and select it
	HRGN hRgnNew;
	HRGN hRgnOutside, hRgnInside;
	hRgnOutside = ::CreateRectRgnIndirect(lpRect);
	RECT rect = *lpRect;
	::InflateRect(&rect, -size.cx, -size.cy);
	::IntersectRect(&rect, &rect, lpRect);
	hRgnInside = ::CreateRectRgnIndirect(&rect);
	hRgnNew = ::CreateRectRgn(0, 0, 0, 0);
	::CombineRgn(hRgnNew, hRgnOutside, hRgnInside, RGN_XOR);

	HBRUSH hBrushOld = NULL;
	if(hBrush == NULL)
		hBrush = CDCHandle::GetHalftoneBrush();
	if(hBrushLast == NULL)
		hBrushLast = hBrush;

	HRGN hRgnLast = NULL, hRgnUpdate = NULL;
	if(lpRectLast != NULL)
	{
		// find difference between new region and old region
		hRgnLast = ::CreateRectRgn(0, 0, 0, 0);
		::SetRectRgn(hRgnOutside, lpRectLast->left, lpRectLast->top, lpRectLast->right, lpRectLast->bottom);
		rect = *lpRectLast;
		::InflateRect(&rect, -sizeLast.cx, -sizeLast.cy);
		::IntersectRect(&rect, &rect, lpRectLast);
		::SetRectRgn(hRgnInside, rect.left, rect.top, rect.right, rect.bottom);
		::CombineRgn(hRgnLast, hRgnOutside, hRgnInside, RGN_XOR);

		// only diff them if brushes are the same
		if(hBrush == hBrushLast)
		{
			hRgnUpdate = ::CreateRectRgn(0, 0, 0, 0);
			::CombineRgn(hRgnUpdate, hRgnLast, hRgnNew, RGN_XOR);
		}
	}
	if(hBrush != hBrushLast && lpRectLast != NULL)
	{
		// brushes are different -- erase old region first
		dc.SelectClipRgn(hRgnLast);
		dc.GetClipBox(&rect);
		hBrushOld = dc.SelectBrush(hBrushLast);
		dc.PatBlt(rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, PATINVERT);
		dc.SelectBrush(hBrushOld);
		hBrushOld = NULL;
	}

	// draw into the update/new region
	dc.SelectClipRgn(hRgnUpdate != NULL ? hRgnUpdate : hRgnNew);
	dc.GetClipBox(&rect);
	hBrushOld = dc.SelectBrush(hBrush);
	dc.PatBlt(rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, PATINVERT);

	// cleanup DC ******************* fixed ************************
	if(hBrushOld != NULL)
		dc.SelectBrush(hBrushOld);
	dc.SelectClipRgn(NULL);
	if(NULL != hBrush)
		DeleteObject(hBrush); //Free our halftone brush
	DeleteObject(hRgnNew);
	DeleteObject(hRgnOutside);
	DeleteObject(hRgnInside);
	if (NULL != hRgnLast)
		DeleteObject(hRgnLast);
	if (NULL != hRgnUpdate)
		DeleteObject(hRgnUpdate);
}


#if 1	//+++ DonutRAPT pN炵Ă[`.
// WinNT3.5+: ̗\̈ꎞIɍŏBEBhEŏꍇƓ
void RtlSetMinProcWorkingSetSize()
{
	HINSTANCE hLib = ::LoadLibrary( _T( "kernel32.dll" ) );
	if ( hLib ) {
		typedef BOOL ( WINAPI *LPFN )( HANDLE, SIZE_T, SIZE_T );
		LPFN fnSetProcessWorkingSetSize = reinterpret_cast< LPFN >( ::GetProcAddress( hLib, "SetProcessWorkingSetSize" ) );
		if ( fnSetProcessWorkingSetSize ) {
			HANDLE hCurrentProcess = ::GetCurrentProcess();
		  #ifdef _WIN64
			MTLVERIFY( fnSetProcessWorkingSetSize(hCurrentProcess, 0xFFFFFFFFFFFFFFFFll, 0xFFFFFFFFFFFFFFFFll ) );
		  #else
			MTLVERIFY( fnSetProcessWorkingSetSize(hCurrentProcess, 0xFFFFFFFF, 0xFFFFFFFF ) );
		  #endif
		}
		::FreeLibrary( hLib );
	}
}
#endif


} // namespace MTL;

#endif // __MTLMISC_H__
